import numpy as np
from model import FlatNet
from loader import FlatData
import torch
import torch.nn as nn
import torch.optim as optim
from scipy import stats
import queue
import random
import copy
class FlatLink():
    """
    Linker for environment.
    Training: Generate training data, accumulate n-step reward and train model.
    Play: Execute trained model.
    """
    def __init__(self,size_x=20,size_y=10, gpu_ = 0):
        self.n_step_return_ = 100
        self.gamma_ = 0.99
        self.batch_size = 4000
        self.num_epochs = 100
        self.play_epochs = 1000
        self.a = 3.57
        self.thres = 1.0
        self.alpha_rv = stats.alpha(self.a)
        self.alpha_rv.random_state = np.random.RandomState(seed=342423)
        self.model = FlatNet(size_x,size_y)
        self.queue = None
        self.max_iter =  self.n_step_return_
        self.dtype = torch.float
        self.play_game_number = 100
        self.rate_decay = 0.8
        self.size_x=size_x
        self.size_y=size_y
        self.replay_buffer_size = 10000
        self.buffer_multiplicator = 1
        self.best_loss = 100000.0
        self.min_thres = 0.03
        self.normalize_reward = 100
        if gpu_>-1:
            self.cuda = True
            self.device = torch.device('cuda', gpu_)
            self.model.cuda()
        else:
            self.cuda = False
            self.device = torch.device('cpu', 0)

    def load_(self, file='model_a3c_i.pt'):
        self.model = torch.load(file)
        print(self.model)
    
    def test_random_move(self):
        move =  np.argmax(self.alpha_rv.rvs(size=5))
        print("move ",move)
   
    def get_action(self, policy):
        """
        Return either predicted (from policy) or random action
        """
        if np.abs(stats.norm(1).rvs()) < self.thres:
            move =  np.argmax(self.alpha_rv.rvs(size=len(policy)))
            return move, 0
        else:
            xk = np.arange(len(policy))
            custm = stats.rv_discrete(name='custm', values=(xk, policy))
            return custm.rvs(), 1

    def accumulate(self, memory, state_value, done_):

        """
        memory has form (state, action, reward, after_state, done)
        Accumulate b-step reward and put to training-queue
        """
        n = min(len(memory), self.n_step_return_)
        curr_state = memory[0][0]
        action =memory[0][1]
        
        n_step_return = 0
        for i in range(n):
            reward  = memory[i][2]
            n_step_return += np.power(self.gamma_,i) * reward
        
        done = memory[-1][4]
        #if done:
        #    print("n_step_return: ",n_step_return," done: ", done)
        
        if not done:
            n_step_return += np.power(self.gamma_,n)*state_value[0]
        else:
            n_step_return += 1.0
        
        if done_:
            n_step_return += 10.0
        #if not done:
        #    n_step_return += (-1)*np.power(self.gamma_,n+1)*np.abs(state_value)[0]
        #else:
        #    n_step_return += np.abs(state_value)[0] # Not neccessary!

        if len(memory)==1 or done:
            memory = []
        else:
            memory = memory[1:-1]
        n_step_return /= self.normalize_reward
        #print("n_step_return: ",n_step_return," n: ", n)
        n_step_return = np.clip(n_step_return, -1., 1.)
        #print("n_step_return: ",n_step_return," 4")
        return curr_state, action, n_step_return , memory

   
    def training(self,observations,targets,rewards):
        """
        Run model training on accumulated training experience
        """
        self.model.train()
        data_train = FlatData(observations,targets,rewards,self.model.num_actions_,self.device, self.dtype)
        dataset_sizes = len(data_train)
        dataloader = torch.utils.data.DataLoader(data_train, batch_size=self.batch_size, shuffle=True,num_workers=0)
        best_model_wts = None
        optimizer= optim.Adam(self.model.parameters(),lr=0.0001,weight_decay=0.02)
        best_loss  = 10000000
        # TODO early stop
        for epoch in range(self.num_epochs):
            running_loss = 0.0
            for observation, obs_2, target, reward in dataloader:
                optimizer.zero_grad()
                policy, value = self.model.forward(observation.to(device = self.device, dtype=self.dtype),obs_2.to(device = self.device, dtype=self.dtype))
                loss = self.model.loss(policy,target.to(device = self.device, dtype=self.dtype),value,reward.to(device = self.device, dtype=self.dtype))
                loss.backward()
                optimizer.step()
                running_loss += torch.abs(loss).item() * observation.size(0)
               
            epoch_loss = running_loss / dataset_sizes
           
            if epoch_loss < self.best_loss:
                best_loss = epoch_loss
                best_model_wts = copy.deepcopy(self.model.state_dict())
                try:
                    torch.save(self.model, 'model_a3c_i.pt')
                    
                except:
                    print("Model is not saved!!!")
        print('Last Loss {}, Best Loss {}'.format(epoch_loss, best_loss))
        if best_model_wts is not None:
            self.model.load_state_dict(best_model_wts)


    def predict(self, observation):
        """
        Forward pass on model
        Returns: Policy, Value
        """
        state = observation[0]
        state2 = observation[1]
        observation1 = state.astype(np.float).reshape((1,state.shape[0],state.shape[1],state.shape[2]))
        observation2 = state2.astype(np.float).reshape((1,state2.shape[0]))
        val1 = torch.from_numpy(observation1.astype(np.float)).to(device = self.device, dtype=self.dtype)
        val2 = torch.from_numpy(observation2.astype(np.float)).to(device = self.device, dtype=self.dtype)
        p, v = self.model.forward(val1,val2)
        return p.cpu().detach().numpy().reshape(-1), v.cpu().detach().numpy().reshape(-1)

    
    def trainer(self,buffer_list = []):
        """
        Call training sequence if buffer not emtpy
        """
        if buffer_list==[]:
            return
        curr_state_, action, n_step_return = zip(*buffer_list)
        self.training( curr_state_, action, n_step_return)
        #self.queue.queue.clear()
        #try:
        #    torch.save(self.model, 'model_a3c.pt')
        #except:
        #    print("Model is not saved!!!")


class FlatConvert(FlatLink):
    def __init__(self,size_x,size_y, gpu_ = 0):
        super(FlatConvert, self).__init__(size_x,size_y,gpu_)


    def perform_training(self,env,env_renderer=None):
        """
        Run simulation in training mode
        """
        self.queue = queue.Queue()
        buffer_list=[]
        best_reward = -999999999
        #self.load_('model_a3c.pt')
        print("start training...")
        number_of_agents = len(env.agents)
        for j in range(self.play_epochs):
            reward_mean= 0.0
            predicted_mean = 0.0
            steps_mean = np.zeros((number_of_agents))
            solved_mean = 0.0
            b_list=[]
            
            for i in range(self.play_game_number):
                predicted_, reward_ ,steps_, done_= self.collect_training_data(env,env_renderer)
                reward_mean+=reward_
                predicted_mean+=predicted_
                solved_mean += done_
                steps_mean += np.floor(steps_)
 
            b_list = list(self.queue.queue) 
            reward_mean /= float(self.play_game_number)
            predicted_mean /= float(self.play_game_number)
            steps_mean /= float(self.play_game_number)
            solved_mean /= float(self.play_game_number)
            print("play epoch: ",j,", total buffer size: ", len(b_list))
            self.trainer(buffer_list + b_list)
            buffer_list =  buffer_list  + b_list
            if len(buffer_list) > self.replay_buffer_size * 1.2:
                list_2 =[ [x,y,z] for x,y,z in sorted(buffer_list, key=lambda pair: pair[2],reverse=True)] 
                size_  = int(self.replay_buffer_size/2)
                buffer_list =  list_2[:size_] + list_2[-size_:]
           
            if reward_mean > best_reward:
                best_reward = reward_mean
                try:
                    torch.save(self.model, 'model_a3c.pt')
                except:
                    print("Model is not saved!!!")

            print("play epoch: ",j,", buffer size: ", len(buffer_list),", reward (mean): ",reward_mean,
            ", solved (mean): ",solved_mean, ", steps (mean): ",steps_mean,", predicted (mean): ",predicted_mean)
           
            self.queue.queue.clear()
            self.thres = max( self.thres*self.rate_decay,self.min_thres)
            env.num_resets = 0
            
    
    def shape_reward(self,states_,reward):
        """
        Additional penalty for not moving (e.g when in deadlock), only applied on last time step!
        """
        #if np.array_equal(states_[0,2,:,:],states_[1,2,:,:]):
        #    reward += -0.5
        #states_=states_[0].reshape((4,int(states_[0].shape[0]/4),states_[0].shape[1],states_[0].shape[2]))
        #x = states_.shape[0]
        #if np.array_equal(states_[x-2,2,:,:],states_[x-1,2,:,:]):
        #reward += -0.5
        return reward

    def make_state(self,state_):
        """
        Stack states 3 times for initial call (t1==t2==t3)
        """
        state = state_[0] #target does not move!
        return [np.stack((state,state,state,state)).reshape((4*state.shape[0],state.shape[1],state.shape[2])),state_[1]]

    def update_state(self,states_a,state):
        """
        update state by removing oldest timestep and adding actual step
        """
        states_ = states_a[0]
        states_=states_.reshape((4,state[0].shape[0],state[0].shape[1],state[0].shape[2]))
        states_[:(states_.shape[0]-1),:,:,:] =  states_[1:,:,:,:]
        states_[(state[0].shape[0]-1):,:,:,:] = state[0].reshape((1,state[0].shape[0],state[0].shape[1],state[0].shape[2]))
        return [states_.reshape((4*state[0].shape[0],state[0].shape[1],state[0].shape[2])),state[1]]


    def stack_observations(self,state):
        local_rail_obs=state[0] 
        obs_map_state=state[1] 
        obs_other_agents_state=state[2] 
        direction=state[3] 
        new_array = np.dstack((local_rail_obs, obs_map_state, obs_other_agents_state))
        new_array = np.transpose( new_array, (2, 0, 1))
        return [new_array,direction]

    def init_state_dict(self,state_dict_):
        for key in state_dict_.keys():
            state_dict_[key] = self.stack_observations(state_dict_[key])
        return state_dict_

    def update_state_dict(self,old_state_dict, new_state_dict):
        for key in old_state_dict.keys():
            new_state_dict[key] = self.stack_observations(new_state_dict[key])
        return new_state_dict

    def play(self,env,env_renderer=None,filename = 'model_a3c.pt',epochs_ =10):
        """
        Run simulation on trained model
        """
        episode_reward = 0
        self.thres = 0.0
        self.load_(filename)
        self.model.eval()
        global_reward = 0
        for i in range(epochs_):
            policy_dict = {}
            value_dict = {}
            action_dict = {}
            #self.max_iter = 40
            iter_ = 0
            pre_ = 0
            act_ = 0
            done = False
            curr_state_dict = self.init_state_dict(env.reset())
            #curr_state_dict = env.reset()
            while not done and iter_< self.max_iter:
                iter_+=1
                for agent_id, state in curr_state_dict.items():
                    policy, value = self.predict(state)
                    policy_dict[agent_id] = policy
                    value_dict[agent_id] = value
                    action_dict[agent_id], pre = self.get_action(policy)
                    pre_ += pre
                    act_+=1 
                if env_renderer is not None:    
                    env_renderer.render_env(show=True)
                next_state_dict, reward_dict, done_dict, _ = env.step(action_dict)
                
                next_state_dict = self.update_state_dict(curr_state_dict, next_state_dict)
                done = done_dict['__all__']
                
                curr_state_dict = next_state_dict

        return  float(pre_)/float(act_), global_reward, float(iter_)/float(self.max_iter)




    def collect_training_data(self,env,env_renderer = None):
        """
        Run single simualtion szenario for training
        """
        done = False
        curr_state_dict = self.init_state_dict(env.reset())
        if env_renderer is not None:
            env_renderer.reset()
        #curr_state_dict = env.reset()
        episode_reward = 0
        memory={}
        for agent_id, state in curr_state_dict.items():
            memory[agent_id] = []
        self.model.eval()
        policy_dict = {}
        value_dict = {}
        action_dict = {}
        iter_counts = [0] * len(curr_state_dict)
        iter_ = 0
        pre_ = 0
        act_ = 0
        global_reward = 0.0
        while not done and iter_< self.max_iter:
            iter_+=1
            for agent_id, state in curr_state_dict.items():
                policy, value = self.predict(state)
                policy_dict[agent_id] = policy
                value_dict[agent_id] = value
                action_dict[agent_id], pre = self.get_action(policy)
                pre_ += pre
                act_+=1 

            next_state_dict, reward_dict, done_dict, _ = env.step(action_dict)
            next_state_dict = self.update_state_dict(curr_state_dict, next_state_dict)
            #reward_dict = self.modify_reward(reward_dict,done_dict,next_state_dict)


            done = done_dict['__all__']
            for agent_id, state in curr_state_dict.items():
                if done_dict[agent_id] is not True:
                    iter_counts[agent_id] += 1
                if memory[agent_id] == [] and done_dict[agent_id]:
                    continue
                memory[agent_id].append(tuple([curr_state_dict[agent_id], 
                action_dict[agent_id], reward_dict[agent_id], next_state_dict[agent_id], done_dict[agent_id]]))
                reward_dict[agent_id] = self.shape_reward(next_state_dict[agent_id], reward_dict[agent_id])
                global_reward += reward_dict[agent_id]
                while ((len(memory[agent_id]) >= self.n_step_return_) or (done_dict[agent_id] and not memory[agent_id] == [])):
                    curr_state, action, n_step_return, memory[agent_id] = self.accumulate(memory[agent_id], 
                    value_dict[agent_id],done)
                    self.queue.put([curr_state, action, n_step_return])


            curr_state_dict = next_state_dict

            if env_renderer is not None:
                #does not work for actual masterè    
                env_renderer.render_env(show=True,show_observations=True)

        return  float(pre_)/float(act_), global_reward, np.array(iter_counts).astype(np.float)/float(self.max_iter), 1 if done else 0